import { JsObjects, Parsing, TsKeywords, Validation, type inferred } from "@arktype/schema";
import type { MatchParser } from "./match.js";
import { Scope, type Module, type ScopeParser } from "./scope.js";
import type { DeclarationParser, DefinitionParser, Generic, TypeParser } from "./type.js";
/** Root scopes can be inferred automatically from node definitions, but
 * explicitly typing them can improve responsiveness */
export type RootScope<exports> = Scope<{
    exports: exports;
    locals: {};
    ambient: {};
}>;
export type ArkResolutions = {
    exports: Ark;
    locals: {};
    ambient: Ark;
};
export type ParsingResolutions = {
    exports: Parsing.infer;
    locals: {};
    ambient: {};
};
export type TsGenericsResolutions<$ = Ark> = {
    exports: TsGenericsExports<$>;
    locals: {};
    ambient: {};
};
type TsGenericsExports<$ = Ark> = {
    Record: Generic<[
        "K",
        "V"
    ], {
        "[K]": "V";
    }, $>;
};
export declare const tsGenerics: Module<TsGenericsResolutions<Ark>>;
export declare const ark: Scope<ArkResolutions>;
export declare const arktypes: Module<ArkResolutions>;
export interface Ark extends TsKeywords.infer, JsObjects.infer, Validation.infer, TsGenericsExports {
    parse: Module<ParsingResolutions>;
}
export declare const scope: ScopeParser<{}, Ark>;
export declare const type: TypeParser<Ark>;
export declare const match: MatchParser<Ark>;
export declare namespace type {
    type cast<to> = {
        [inferred]?: to;
    };
}
export declare const define: DefinitionParser<Ark>;
export declare const declare: DeclarationParser<Ark>;
export {};
//# sourceMappingURL=ark.d.ts.map