import { isNode, schema } from "@arktype/schema";
import { isThunk, objectKindOf, printable, throwParseError } from "@arktype/util";
import { Type } from "../type.js";
import { parseObjectLiteral } from "./objectLiteral.js";
import { parseTuple } from "./tuple.js";
export const parseObject = (def, ctx) => {
    const objectKind = objectKindOf(def);
    switch (objectKind) {
        case undefined:
            if (def instanceof Type) {
                return def.root;
            }
            if (isNode(def) && def.isType()) {
                return def;
            }
            return parseObjectLiteral(def, ctx);
        case "Array":
            return parseTuple(def, ctx);
        case "RegExp":
            return schema({ basis: "string", pattern: def });
        case "Function":
            const resolvedDef = isThunk(def) ? def() : def;
            if (resolvedDef instanceof Type) {
                return resolvedDef.root;
            }
            return throwParseError(writeBadDefinitionTypeMessage("Function"));
        default:
            return throwParseError(writeBadDefinitionTypeMessage(objectKind ?? printable(def)));
    }
};
export const writeBadDefinitionTypeMessage = (actual) => `Type definitions must be strings or objects (was ${actual})`;
//# sourceMappingURL=definition.js.map