import { type TypeNode } from "@arktype/schema";
import { type Dict, type ErrorMessage, type evaluate, type merge } from "@arktype/util";
import type { ParseContext } from "../scope.js";
import type { inferDefinition, validateDefinition } from "./definition.js";
import type { astToString } from "./semantic/utils.js";
import type { validateString } from "./semantic/validate.js";
import { parseEntry, type EntryParseResult, type IndexedKey, type OptionalValue, type validateObjectValue } from "./shared.js";
export declare const parseObjectLiteral: (def: Dict, ctx: ParseContext) => TypeNode;
/**
 * Infers the contents of an object literal, ignoring a spread definition
 * You probably want to use {@link inferObjectLiteral} instead.
 */
type inferObjectLiteralInner<def extends object, $, args> = {
    -readonly [k in keyof def as nonOptionalKeyFrom<k, def[k], $, args>]: inferDefinition<def[k], $, args>;
} & {
    -readonly [k in keyof def as optionalKeyFrom<k, def[k]>]?: inferDefinition<def[k] extends OptionalValue<infer inner> ? inner : def[k], $, args>;
};
export type inferObjectLiteral<def extends object, $, args> = evaluate<"..." extends keyof def ? merge<inferDefinition<def["..."], $, args>, inferObjectLiteralInner<def, $, args>> : inferObjectLiteralInner<def, $, args>>;
export type validateObjectLiteral<def, $, args> = {
    [k in keyof def]: k extends IndexedKey<infer indexDef> ? validateString<indexDef, $, args> extends ErrorMessage<infer message> ? indexParseError<message> : inferDefinition<indexDef, $, args> extends PropertyKey ? validateDefinition<def[k], $, args> : indexParseError<writeInvalidPropertyKeyMessage<indexDef>> : k extends "..." ? inferDefinition<def[k], $, args> extends object ? validateDefinition<def[k], $, args> : indexParseError<writeInvalidSpreadTypeMessage<astToString<def[k]>>> : validateObjectValue<def[k], $, args>;
};
type nonOptionalKeyFrom<k extends PropertyKey, valueDef, $, args> = parseEntry<k, valueDef> extends infer result extends EntryParseResult ? result["kind"] extends "required" ? result["innerKey"] : result["kind"] extends "indexed" ? inferDefinition<result["innerKey"], $, args> & PropertyKey : never : never;
type optionalKeyFrom<k extends PropertyKey, valueDef> = parseEntry<k, valueDef> extends infer result extends EntryParseResult<"optional"> ? result["innerKey"] : never;
declare const indexParseSymbol: unique symbol;
export type indexParseError<message extends string = string> = {
    [indexParseSymbol]: message;
};
export declare const writeInvalidPropertyKeyMessage: <indexDef extends string>(indexDef: indexDef) => `Indexed key definition '${indexDef}' must be a string, number or symbol`;
type writeInvalidPropertyKeyMessage<indexDef extends string> = `Indexed key definition '${indexDef}' must be a string, number or symbol`;
export declare const writeInvalidSpreadTypeMessage: <def extends string>(def: def) => `Spread operand must resolve to an object literal type (was ${def})`;
type writeInvalidSpreadTypeMessage<def extends string> = `Spread operand must resolve to an object literal type (was ${def})`;
export {};
//# sourceMappingURL=objectLiteral.d.ts.map