import type { BaseMeta, Node } from "@arktype/schema";
import type { evaluate } from "@arktype/util";
import type { validateDefinition } from "./definition.js";
import { Scanner } from "./string/shift/scanner.js";
type parsedKey<result extends KeyParseResult> = result;
type OptionalTuple<value> = readonly [value, "?"];
type KeyParseResult<kind extends ParsedKeyKind = ParsedKeyKind> = {
    kind: kind;
    innerKey: string | symbol;
};
export type EntryParseResult<kind extends ParsedKeyKind = ParsedKeyKind> = evaluate<KeyParseResult<kind> & {
    innerValue: unknown;
}>;
type ParsedKeyKind = "required" | "optional" | "indexed" | "spread";
type parsedEntry<result extends EntryParseResult> = result;
export type OptionalValue<value> = OptionalStringDefinition<value & string> | OptionalTuple<value>;
export type OptionalStringDefinition<name extends string = string> = `${name}?`;
export type IndexedKey<def extends string = string> = `[${def}]`;
export type validateObjectValue<def, $, args> = def extends OptionalTuple<infer value> ? readonly [validateObjectValueString<value, $, args>, "?"] : validateObjectValueString<def, $, args>;
type validateObjectValueString<def, $, args> = def extends OptionalStringDefinition<infer innerValue> ? `${validateDefinition<innerValue, $, args>}?` : validateDefinition<def, $, args>;
type DefinitionEntry = readonly [string | symbol, unknown];
export declare const parseEntry: ([key, value]: DefinitionEntry) => EntryParseResult;
export type parseEntry<keyDef extends PropertyKey, valueDef> = parseKey<keyDef> extends infer keyParseResult extends KeyParseResult ? valueDef extends OptionalValue<infer innerValue> ? parsedEntry<{
    kind: "optional";
    innerKey: keyParseResult["innerKey"];
    innerValue: innerValue extends OptionalStringDefinition<infer innerStringValue> ? innerStringValue : innerValue;
}> : parsedEntry<{
    kind: keyParseResult["kind"];
    innerKey: keyParseResult["innerKey"];
    innerValue: valueDef;
}> : never;
type parseKey<k> = k extends OptionalStringDefinition<infer inner> ? inner extends `${infer baseName}${Scanner.EscapeToken}` ? parsedKey<{
    kind: "required";
    innerKey: OptionalStringDefinition<baseName>;
}> : parsedKey<{
    kind: "optional";
    innerKey: inner;
}> : k extends "..." ? parsedKey<{
    kind: "spread";
    innerKey: "...";
}> : k extends "\\..." ? parsedKey<{
    kind: "required";
    innerKey: "...";
}> : k extends IndexedKey<infer def> ? parsedKey<{
    kind: "indexed";
    innerKey: def;
}> : k extends `${Scanner.EscapeToken}${infer escapedIndexKey extends IndexedKey}` ? parsedKey<{
    kind: "required";
    innerKey: escapedIndexKey;
}> : parsedKey<{
    kind: "required";
    innerKey: k & (string | symbol);
}>;
export declare const configureShallowDescendants: <node extends Node>(node: node, configOrDescription: BaseMeta | string) => node;
export {};
//# sourceMappingURL=shared.d.ts.map