import { Scanner } from "./string/shift/scanner.js";
const getInnerValue = (value) => {
    if (typeof value === "string") {
        if (value[value.length - 1] === "?") {
            return {
                kind: "optional",
                innerValue: value.slice(0, -1)
            };
        }
    }
    else if (Array.isArray(value)) {
        if (value.length === 2 && value[1] === "?") {
            return {
                kind: "optional",
                innerValue: getInnerValue(value[0]).innerValue
            };
        }
    }
    return {
        kind: "required",
        innerValue: value
    };
};
// these methods are also used in tuple parsing, however the keys of a tuple will always be 0, 1,
// 2, etc and never be `...`, meaning `"spread"` as a kind will never occur in tuples
export const parseEntry = ([key, value]) => {
    const keyParseResult = typeof key === "string" && key.at(-1) === "?"
        ? key.at(-2) === Scanner.escapeToken
            ? { innerKey: `${key.slice(0, -2)}?`, kind: "required" }
            : { innerKey: key.slice(0, -1), kind: "optional" }
        : key === "..."
            ? { innerKey: "...", kind: "spread" }
            : key === "\\..."
                ? { innerKey: "...", kind: "required" }
                : { innerKey: key, kind: "required" };
    const valueParseResult = getInnerValue(value);
    return {
        innerKey: keyParseResult.innerKey,
        innerValue: valueParseResult.innerValue,
        kind: keyParseResult.kind === "indexed"
            ? "indexed"
            : valueParseResult.kind === "optional"
                ? "optional"
                : keyParseResult.kind
    };
};
export const configureShallowDescendants = (node, configOrDescription) => {
    const config = typeof configOrDescription === "string"
        ? { description: configOrDescription }
        : configOrDescription;
    return node.transform((kind, inner) => ({ ...inner, ...config }), 
    // TODO: change to props
    (node) => node.kind !== "required" && node.kind !== "optional");
};
//# sourceMappingURL=shared.js.map