import { type TypeNode } from "@arktype/schema";
import { type BigintLiteral, type Completion, type ErrorMessage, type join, type NumberLiteral } from "@arktype/util";
import type { Module } from "../../../../scope.js";
import type { Generic, GenericProps } from "../../../../type.js";
import type { GenericInstantiationAst } from "../../../semantic/semantic.js";
import type { DynamicState } from "../../reduce/dynamic.js";
import type { state, StaticState } from "../../reduce/static.js";
import type { BaseCompletions } from "../../string.js";
import type { Scanner } from "../scanner.js";
import { parseGenericArgs, writeInvalidGenericArgsMessage, type ParsedArgs } from "./genericArgs.js";
export declare const parseUnenclosed: (s: DynamicState) => void;
export type parseUnenclosed<s extends StaticState, $, args> = Scanner.shiftUntilNextTerminator<s["unscanned"]> extends Scanner.shiftResult<infer token, infer unscanned> ? token extends "keyof" ? state.addPrefix<s, "keyof", unscanned> : tryResolve<s, token, $, args> extends infer result ? result extends ErrorMessage<infer message> ? state.error<message> : result extends keyof $ ? $[result] extends GenericProps ? parseGenericInstantiation<token, $[result], state.scanTo<s, unscanned>, $, args> : state.setRoot<s, result, unscanned> : state.setRoot<s, result, unscanned> : never : never;
export declare const parseGenericInstantiation: (name: string, g: Generic, s: DynamicState) => TypeNode;
export type parseGenericInstantiation<name extends string, g extends GenericProps, s extends StaticState, $, args> = Scanner.skipWhitespace<s["unscanned"]> extends `<${infer unscanned}` ? parseGenericArgs<name, g["parameters"], unscanned, $, args> extends infer result ? result extends ParsedArgs<infer argAsts, infer nextUnscanned> ? state.setRoot<s, GenericInstantiationAst<g, argAsts>, nextUnscanned> : result : never : state.error<writeInvalidGenericArgsMessage<name, g["parameters"], []>>;
type tryResolve<s extends StaticState, token extends string, $, args> = token extends keyof $ ? token : token extends keyof args ? token : token extends NumberLiteral ? token : token extends BigintLiteral ? token : token extends `${infer submodule extends keyof $ & string}.${infer reference}` ? $[submodule] extends Module<infer r> ? reference extends keyof r["exports"] ? token : unknown extends r["exports"] ? ErrorMessage<writeNonSubmoduleDotMessage<submodule>> : unresolvableError<s, reference, $[submodule], args, [submodule]> : ErrorMessage<writeNonSubmoduleDotMessage<submodule>> : unresolvableError<s, token, $, args, []>;
export declare const writeNonSubmoduleDotMessage: <name extends string>(name: name) => `'${name}' must reference a module to be accessed using dot syntax`;
type writeNonSubmoduleDotMessage<name extends string> = `'${name}' must reference a module to be accessed using dot syntax`;
export declare const writeMissingSubmoduleAccessMessage: <name extends string>(name: name) => `Reference to submodule '${name}' must specify an alias`;
export type writeMissingSubmoduleAccessMessage<name extends string> = `Reference to submodule '${name}' must specify an alias`;
/** Provide valid completions for the current token, or fallback to an
 * unresolvable error if there are none */
export type unresolvableError<s extends StaticState, token extends string, $, args, submodulePath extends string[]> = validReferenceFromToken<token, $, args, submodulePath> extends never ? ErrorMessage<writeUnresolvableMessage<qualifiedReference<token, submodulePath>>> : Completion<`${s["scanned"]}${qualifiedReference<validReferenceFromToken<token, $, args, submodulePath>, submodulePath>}`>;
type qualifiedReference<reference extends string, submodulePath extends string[]> = join<[...submodulePath, reference], ".">;
type validReferenceFromToken<token extends string, $, args, submodulePath extends string[]> = Extract<submodulePath extends [] ? BaseCompletions<$, args> : keyof $, `${token}${string}`>;
export declare const writeUnresolvableMessage: <token extends string>(token: token) => `'${token}' is unresolvable`;
type writeUnresolvableMessage<token extends string> = `'${token}' is unresolvable`;
export declare const writeMissingOperandMessage: <s extends DynamicState>(s: s) => string;
export type writeMissingOperandMessage<s extends StaticState, operator extends string | undefined = state.previousOperator<s>> = operator extends string ? writeMissingRightOperandMessage<operator, s["unscanned"]> : writeExpressionExpectedMessage<s["unscanned"]>;
export type writeMissingRightOperandMessage<token extends string, unscanned extends string = ""> = `Token '${token}' requires a right operand${unscanned extends "" ? "" : ` before '${unscanned}'`}`;
export declare const writeMissingRightOperandMessage: <token extends string, unscanned extends string>(token: token, unscanned?: unscanned) => `Token '${token}' requires a right operand${unscanned extends "" ? "" : ` before '${unscanned}'`}`;
export declare const writeExpressionExpectedMessage: <unscanned extends string>(unscanned: unscanned) => string;
export type writeExpressionExpectedMessage<unscanned extends string> = `Expected an expression${unscanned extends "" ? "" : ` before '${unscanned}'`}`;
export {};
//# sourceMappingURL=unenclosed.d.ts.map