import { type ArkConfig, type TypeNode, type extractIn, type extractOut } from "@arktype/schema";
import { type Dict, type evaluate, type isAny, type nominal } from "@arktype/util";
import type { type } from "./ark.js";
import { type MatchParser } from "./match.js";
import { type inferDefinition, type validateDefinition } from "./parser/definition.js";
import { parseGenericParams, type GenericDeclaration, type GenericParamsParseError } from "./parser/generic.js";
import { Type, type DeclarationParser, type DefinitionParser, type Generic, type GenericProps, type TypeParser } from "./type.js";
import { type arkKind } from "./util.js";
export type ScopeParser<parent, ambient> = {
    <const def>(def: validateScope<def, parent & ambient>): Scope<{
        exports: inferBootstrapped<{
            exports: bootstrapExports<def>;
            locals: bootstrapLocals<def> & parent;
            ambient: ambient;
        }>;
        locals: inferBootstrapped<{
            exports: bootstrapLocals<def>;
            locals: bootstrapExports<def> & parent;
            ambient: ambient;
        }>;
        ambient: ambient;
    }>;
};
export interface ArkScopeConfig extends ArkConfig {
    ambient?: Scope | null;
}
type validateScope<def, $> = {
    [k in keyof def]: parseScopeKey<k>["params"] extends [] ? def[k] extends Type | PreparsedResolution ? def[k] : validateDefinition<def[k], $ & bootstrap<def>, {}> : parseScopeKey<k>["params"] extends GenericParamsParseError ? parseScopeKey<k>["params"][0] : validateDefinition<def[k], $ & bootstrap<def>, {
        [param in parseScopeKey<k>["params"][number]]: unknown;
    }>;
};
export type bindThis<def> = {
    this: Def<def>;
};
/** nominal type for an unparsed definition used during scope bootstrapping */
type Def<def = {}> = nominal<def, "unparsed">;
/** sentinel indicating a scope that will be associated with a generic has not yet been parsed */
export type UnparsedScope = "$";
type bootstrap<def> = bootstrapLocals<def> & bootstrapExports<def>;
type bootstrapLocals<def> = bootstrapAliases<{
    [k in keyof def & PrivateDeclaration as extractPrivateKey<k>]: def[k];
}>;
type bootstrapExports<def> = bootstrapAliases<{
    [k in Exclude<keyof def, PrivateDeclaration>]: def[k];
}>;
/** These are legal as values of a scope but not as definitions in other contexts */
type PreparsedResolution = Module | GenericProps;
type bootstrapAliases<def> = {
    [k in Exclude<keyof def, GenericDeclaration | symbol>]: def[k] extends PreparsedResolution ? def[k] : def[k] extends (() => infer thunkReturn extends PreparsedResolution) ? thunkReturn : Def<def[k]>;
} & {
    [k in keyof def & GenericDeclaration as extractGenericName<k>]: Generic<parseGenericParams<extractGenericParameters<k>>, def[k], UnparsedScope>;
};
type inferBootstrapped<r extends Resolutions> = evaluate<{
    [name in keyof r["exports"]]: r["exports"][name] extends Def<infer def> ? inferDefinition<def, $<r>, {}> : r["exports"][name] extends GenericProps<infer params, infer def> ? Generic<params, def, $<r>> : r["exports"][name];
}>;
type extractGenericName<k> = k extends GenericDeclaration<infer name> ? name : never;
type extractGenericParameters<k> = k extends GenericDeclaration<string, infer params> ? params : never;
type extractPrivateKey<k> = k extends PrivateDeclaration<infer key> ? key : never;
type PrivateDeclaration<key extends string = string> = `#${key}`;
export type resolve<reference extends keyof $ | keyof args, $, args> = (reference extends keyof args ? args[reference] : $[reference & keyof $]) extends infer resolution ? [resolution] extends [never] ? never : isAny<resolution> extends true ? any : resolution extends Def<infer def> ? inferDefinition<def, $, args> : resolution : never;
export type moduleKeyOf<$> = {
    [k in keyof $]: $[k] extends Module ? k & string : never;
}[keyof $];
export type tryInferSubmoduleReference<$, token> = token extends `${infer submodule extends moduleKeyOf<$>}.${infer subalias}` ? subalias extends keyof $[submodule] ? $[submodule][subalias] extends Type<infer t> ? t : never : never : never;
type $<r extends Resolutions> = r["exports"] & r["locals"] & r["ambient"];
type exportedName<r extends Resolutions> = keyof r["exports"] & string;
export type Module<r extends Resolutions = any> = {
    [k in exportedName<r> | arkKind]: k extends string ? [r["exports"][k]] extends [never] ? Type<never, $<r>> : isAny<r["exports"][k]> extends true ? Type<any, $<r>> : r["exports"][k] extends PreparsedResolution ? r["exports"][k] : Type<r["exports"][k], $<r>> : type.cast<"module">;
};
export type Resolutions = {
    exports: unknown;
    locals: unknown;
    ambient: unknown;
};
export type ParseContext = {
    baseName: string;
    path: string[];
    scope: Scope;
    args: Record<string, TypeNode> | undefined;
};
type ParseContextInput = Pick<ParseContext, "baseName" | "args">;
export declare class Scope<r extends Resolutions = any> {
    infer: extractOut<r["exports"]>;
    inferIn: extractIn<r["exports"]>;
    config: ArkScopeConfig;
    private parseCache;
    private resolutions;
    /** The set of names defined at the root-level of the scope mapped to their
     * corresponding definitions.**/
    aliases: Record<string, unknown>;
    private exportedNames;
    private ambient;
    constructor(def: Dict, config: ArkScopeConfig);
    static root: ScopeParser<{}, {}>;
    type: TypeParser<$<r>>;
    match: MatchParser<$<r>>;
    declare: DeclarationParser<$<r>>;
    scope: ScopeParser<r["exports"], r["ambient"]>;
    define: DefinitionParser<$<r>>;
    toAmbient(): Scope<{
        exports: r["exports"];
        locals: r["locals"];
        ambient: r["exports"];
    }>;
    get<name extends keyof r["exports"] & string>(name: name): Type<r["exports"][name], $<r>>;
    private createRootContext;
    parseDefinition(def: unknown, input: ParseContextInput): TypeNode;
    parse(def: unknown, ctx: ParseContext): TypeNode;
    parseString(def: string, ctx: ParseContext): TypeNode;
    maybeResolve(name: string): TypeNode | Generic | undefined;
    /** If name is a valid reference to a submodule alias, return its resolution  */
    private maybeResolveSubalias;
    maybeResolveNode(name: string): TypeNode | undefined;
    import<names extends exportedName<r>[]>(...names: names): destructuredImportContext<r, names extends [] ? keyof r["exports"] & string : names[number]>;
    bindThis(): {
        this: import("@arktype/schema").IntersectionNode<unknown>;
    };
    private exportedResolutions;
    private exportCache;
    export<names extends exportedName<r>[]>(...names: names): Module<names extends [] ? r : destructuredExportContext<r, names[number]>>;
}
type destructuredExportContext<r extends Resolutions, name extends exportedName<r>> = {
    exports: {
        [k in name]: r["exports"][k];
    };
    locals: r["locals"] & {
        [k in Exclude<keyof r["exports"], name>]: r["exports"][k];
    };
    ambient: r["ambient"];
};
type destructuredImportContext<r extends Resolutions, name extends exportedName<r>> = {
    [k in name as `#${k & string}`]: type.cast<r["exports"][k]>;
};
export declare const writeShallowCycleErrorMessage: (name: string, seen: string[]) => string;
export declare const writeDuplicateNameMessage: <name extends string>(name: name) => `Duplicate name '${name}'`;
type writeDuplicateNameMessage<name extends string> = `Duplicate name '${name}'`;
export type ParsedScopeKey = {
    isLocal: boolean;
    name: string;
    params: string[];
};
export declare const parseScopeKey: (k: string) => ParsedScopeKey;
type parseScopeKey<k> = k extends PrivateDeclaration<infer inner> ? parsePossibleGenericDeclaration<inner, true> : parsePossibleGenericDeclaration<k, false>;
type parsePossibleGenericDeclaration<k, isLocal extends boolean> = k extends GenericDeclaration<infer name, infer paramString> ? {
    isLocal: isLocal;
    name: name;
    params: parseGenericParams<paramString>;
} : {
    isLocal: isLocal;
    name: k;
    params: [];
};
export {};
//# sourceMappingURL=scope.d.ts.map