import { inferred, type ArkResult, type BaseMeta, type Morph, type Out, type Predicate, type TypeNode, type distill, type extractIn, type extractOut, type includesMorphs, type inferMorphOut, type inferNarrow } from "@arktype/schema";
import { Callable, type Constructor, type Json, type conform } from "@arktype/util";
import type { inferDefinition, validateDeclared, validateDefinition } from "./parser/definition.js";
import { parseGenericParams, type GenericParamsParseError } from "./parser/generic.js";
import type { inferIntersection } from "./parser/semantic/intersections.js";
import type { IndexOneOperator, IndexZeroOperator, TupleInfixOperator } from "./parser/tuple.js";
import type { Scope, bindThis } from "./scope.js";
import { arkKind } from "./util.js";
export type TypeParser<$> = {
    <const def>(def: validateTypeRoot<def, $>): Type<inferTypeRoot<def, $>, $>;
    <const zero, const one, const rest extends readonly unknown[]>(_0: zero extends IndexZeroOperator ? zero : validateTypeRoot<zero, $>, _1: zero extends "keyof" ? validateTypeRoot<one, $> : zero extends "instanceof" ? conform<one, Constructor> : zero extends "===" ? conform<one, unknown> : conform<one, IndexOneOperator>, ..._2: zero extends "===" ? rest : zero extends "instanceof" ? conform<rest, readonly Constructor[]> : one extends TupleInfixOperator ? one extends ":" ? [Predicate<extractIn<inferTypeRoot<zero, $>>>] : one extends "=>" ? [
        Morph<extractOut<inferTypeRoot<zero, $>>, unknown>
    ] : one extends "@" ? [string | BaseMeta] : [validateTypeRoot<rest[0], $>] : []): Type<inferTypeRoot<[zero, one, ...rest], $>, $>;
    <params extends string, const def>(params: `<${validateParameterString<params>}>`, def: validateDefinition<def, $, {
        [param in parseGenericParams<params>[number]]: unknown;
    }>): Generic<parseGenericParams<params>, def, $>;
};
export type DeclarationParser<$> = <preinferred>() => {
    type: <def>(def: validateDeclared<preinferred, def, $, bindThis<def>>) => Type<preinferred, $>;
};
export declare const createTypeParser: <$>(scope: Scope) => TypeParser<$>;
export type DefinitionParser<$> = <def>(def: validateDefinition<def, $, bindThis<def>>) => def;
export declare class Type<t = unknown, $ = any> extends Callable<(data: unknown) => ArkResult<distill<extractOut<t>>>> {
    definition: unknown;
    scope: Scope;
    [inferred]: t;
    infer: distill<extractOut<t>>;
    root: TypeNode<t>;
    allows: this["root"]["allows"];
    description: string;
    json: Json;
    constructor(definition: unknown, scope: Scope);
    configure(configOrDescription: BaseMeta | string): this;
    describe(description: string): this;
    from(literal: this["in"]["infer"]): this["in"]["infer"];
    and<def>(def: validateTypeRoot<def, $>): Type<inferIntersection<t, inferTypeRoot<def, $>>, $>;
    or<def>(def: validateTypeRoot<def, $>): Type<t | inferTypeRoot<def, $>, $>;
    morph<morph extends Morph<this["infer"]>>(morph: morph): Type<(In: this["in"]["infer"]) => Out<inferMorphOut<ReturnType<morph>>>, $>;
    morph<morph extends Morph<this["infer"]>, def>(morph: morph, outValidator: validateTypeRoot<def, $>): Type<(In: this["in"]["infer"]) => Out<extractOut<inferTypeRoot<def, $>>>, $>;
    narrow<def extends Predicate<extractOut<t>>>(def: def): Type<includesMorphs<t> extends true ? (In: this["in"]["infer"]) => Out<inferNarrow<this["infer"], def>> : inferNarrow<this["infer"], def>, $>;
    array(): Type<t[], $>;
    keyof(): Type<keyof this["in"]["infer"], $>;
    assert(data: unknown): this["infer"];
    equals<def>(other: validateTypeRoot<def, $>): this is Type<inferTypeRoot<def, $>, $>;
    extends<def>(other: validateTypeRoot<def, $>): this is Type<inferTypeRoot<def, $>, $>;
    private inCache?;
    get in(): Type<extractIn<t>, $>;
    outCache?: Type<extractOut<t>, $>;
    get out(): Type<extractOut<t>, $>;
    toString(): string;
}
export type validateTypeRoot<def, $> = validateDefinition<def, $, bindThis<def>>;
export type inferTypeRoot<def, $> = inferDefinition<def, $, bindThis<def>>;
type validateParameterString<params extends string> = parseGenericParams<params> extends GenericParamsParseError<infer message> ? message : params;
export declare const validateUninstantiatedGeneric: (g: Generic) => Generic<string[], unknown, any>;
export declare const generic: (parameters: string[], definition: unknown, scope: Scope) => Generic<string[], unknown, any>;
export type GenericProps<params extends string[] = string[], def = unknown, $ = any> = {
    [arkKind]: "generic";
    $: $;
    parameters: params;
    definition: def;
    scope: Scope;
};
export type BoundArgs = Record<string, TypeNode>;
export type Generic<params extends string[] = string[], def = unknown, $ = any> = {
    <args>(...args: conform<args, {
        [i in keyof params]: validateTypeRoot<args[i & keyof args], $>;
    }>): Type<inferDefinition<def, $, bindGenericInstantiation<params, $, args>>, $>;
} & GenericProps<params, def, $>;
type bindGenericInstantiation<params extends string[], $, args> = {
    [i in keyof params & `${number}` as params[i]]: inferTypeRoot<args[i & keyof args], $>;
};
export {};
//# sourceMappingURL=type.d.ts.map